module Main(main) where

import Array
import Char
import IO
import List
import Maybe
import Monad
import System
import Time
import Random(randomIO)

import Concurrent
import Control.Exception
import IOExts
import System.Console.GetOpt
import Posix
import Data.Unique

import ConfigFile
import Curses
import Error
import ExampleConf
import Filter
import Gale
import GenUtil
import Version

import UTF8
import Word(Word8)


data Env = Env { 
    envAction :: IO (),
    envVerbose :: Bool, 
    envConfig :: (Maybe String), 
    envJustArgs :: Bool
    }


{-# NOINLINE screenLock #-}
screenLock = unsafePerformIO $ newMVar ()

withLock :: MVar () -> IO a -> IO a
withLock l action = Control.Exception.bracket (takeMVar l) (putMVar l) (const action)

galeFile :: String -> IO String 
galeFile fn = do
    e <- getEnv "HOME"
    return (e ++ "/.gale/" ++ fn)

getEditor = configLookupElse "EDITOR" (getEnvElse "VISUAL" (getEnvElse "EDITOR" (return "vi"))) 

getTmpFile = do
    (v::Int) <- randomIO 
    u <- newUnique
    return ("/tmp/ginsu.puff." ++ show v ++ "." ++ show (hashUnique u))
    
getEnvElse :: String -> IO String -> IO String
getEnvElse var action = handleMost (\_ -> action) $ getEnv var

ioElse :: IO a -> IO a -> IO a
ioElse a b = tryMost a >>= \x -> case x of 
    Left _ -> b
    Right x -> return x

configLookupElse :: String -> IO String -> IO String
configLookupElse v action = configLookup v >>= \x -> case x of
    Just v -> return v
    Nothing -> action
    
    
main = do
    (flags, acats) <- ginsuOpts
    let cats = nub (stdLookupList "GALE_SUBSCRIBE" ++ acats)
    galeFile "ginsu.errorlog" >>= \logname -> openLog logname
    getClockTime >>= \ct -> putLogLn ("Ginsu Starting:" ++ show ct)
    gc <- newGaleContext cats
    Control.Exception.bracket_ initCurses endWin $ do
    setupCharset
    withLock screenLock (cursset Invisible >> erase >> refresh)
    ic <- newChan
    pl <- forkIO $ puffLoop ic gc
    gl <- forkIO $ getchLoop ic
    yo <- newIORef (0::Int)
    plog <- galeFile "ginsu.pufflog"
    ops <- readPuffs plog
    ps <- newIORef ((reverse $ zip [2..] (reverse ops)) ++[(1,Message $ "Welcome to Ginsu! " ++ helpText)])
    cache_render_r <- newIORef ([],-1)
    cache_filter_r <- newIORef []
    mark_set_r <- newIORef (listArray (0,9) (replicate 10 []))
    mapM_ (\n -> checkMark n >>= \f -> modifyIORef mark_set_r f) [1..9]
    next_r <- newIORef (length ops + 5)
    selected_r <- newIORef 1
    filter_r <- newIORef []
    message_r <- newIORef helpText
    mainLoop gc ic yo ps selected_r filter_r next_r message_r cache_render_r cache_filter_r mark_set_r
    destroyGaleContext gc
    killThread pl
    killThread gl
    ps <- liftM (filter selectable . map snd) $ readIORef ps
    plsize <- fmap read $ configLookupElse "PUFFLOG_SIZE" (return "100")
    writePuffs plog (take plsize ps)


toAscii :: String -> [Word8]
toAscii s = map f s where
    f x | ord x > 127 = fromIntegral (ord '?')
    f x = fromIntegral $ ord x

toLatin1 :: String -> [Word8]
toLatin1 s = map f s where
    f x | ord x > 255 = fromIntegral (ord '?')
    f x = fromIntegral $ ord x

setupCharset :: IO ()
setupCharset = fmap (fmap (filter (`notElem` "-_") . map toUpper)) (configLookup "CHARSET") >>= \x -> case x of
    Just "UTF8" -> setCursesOutConv toUTF
    Just "ASCII" -> setCursesOutConv toAscii
    Just "LATIN1" -> setCursesOutConv toLatin1
    Just "ISO88591" -> setCursesOutConv toLatin1
    _ -> return ()

checkMark n = do
    v <- configLookup ("MARK_" ++ show n)
    return $ maybe id id (v >>= \x -> parseFilter x >>= \y -> return (// [(n - 1,[y])]))


puffLoop ic gc = do
    p <- galeNextPuff gc
    writeChan ic $ Right p
    puffLoop ic gc
    
getchLoop ic = getCh >>= \x -> writeChan ic (Left x) >> getchLoop ic

selectable :: Puff -> Bool
selectable (Puff {}) = True
selectable _ = False

apHead f (x:xs) = f x:xs
apHead _ [] = []


mainLoop gc ic yor psr selected_r filter_r next_r message_r cache_render_r cache_filter_r mark_set_r = getKey redraw >>= \x -> case x of
    (KeyChar '\DC2') -> reconnectGaleContext gc >> continue
    (KeyChar 'q') -> return ()
    (KeyChar 'Q') -> return ()
    KeyEnter -> scrollPuffs 1 >> rcontinue
    (KeyChar '\n') -> scrollPuffs 1 >> rcontinue
    (KeyChar '\r') -> scrollPuffs 1 >> rcontinue
    KeyBackspace -> scrollPuffs (-1) >> rcontinue
    (KeyChar '\b')-> scrollPuffs (-1) >> rcontinue
    (KeyChar '\ENQ') -> scrollPuffs 1  >> rcontinue
    (KeyChar '\EM') -> scrollPuffs (-1)  >> rcontinue
    (KeyChar '\EOT') -> scrSize >>= \(ys,_) ->  scrollPuffs (ys `div` 2) >> rcontinue
    (KeyChar '\NAK') -> scrSize >>= \(ys,_) ->  scrollPuffs (- (ys `div` 2)) >> rcontinue
    (KeyChar '\ACK') -> scrSize >>= \(ys,_) ->  scrollPuffs (ys) >> rcontinue
    (KeyChar '\STX') -> scrSize >>= \(ys,_) ->  scrollPuffs (-ys) >> rcontinue
    KeyNPage -> scrSize >>= \(ys,_) -> scrollPuffs (ys) >> rcontinue
    KeyPPage -> scrSize >>= \(ys,_) -> scrollPuffs (-ys) >> rcontinue
    (KeyChar '\f') -> touchWin stdScr >> rcontinue
    KeyResize -> rcontinue
    (KeyChar 'E') -> do
	gc <- galeFile "ginsu.config"
	e <- getEditor
	def_prog_mode
	endWin
	mySystem (e ++ " " ++ gc)
	reset_prog_mode
	loadStdConfiguration ".gale/ginsu.config"
	rcontinue
    (KeyChar 'g') -> writeIORef selected_r 1 >> select_perhaps select_next >> writeIORef yor 0 >> rcontinue 
    (KeyChar 'G') -> do
	ps <- readIORef psr
	case ps of 
	    ((n,_):_) -> writeIORef selected_r n >> select_perhaps select_prev >> scroll_end >> rcontinue 
	    _ -> continue
    (KeyChar 'j') -> select_next >> center_puff >> rcontinue
    (KeyChar 'k') -> select_prev >> center_puff >> rcontinue
    KeyDown -> select_next >> center_puff >> rcontinue
    KeyUp -> select_prev >> center_puff >> rcontinue
    (KeyChar 'u') -> modifyFilter  (drop 1) >> rcontinue
    (KeyChar 'U') -> modifyFilter (const []) >> rcontinue
    (KeyChar '!') -> modifyFilter (apHead f) >> rcontinue where
	f (FilterNot x) = x
	f x = FilterNot x
    (KeyChar 'x') -> modifyFilter f >> rcontinue where
	f (x:y:r) = y:x:r
	f x = x
    (KeyChar 't') -> filter_thread >> rcontinue
    (KeyChar 'T') -> filter_thread >> rcontinue
    (KeyChar 'm') -> mcontinue
    (KeyChar 'a') -> do
	p <- sel_puff
	case p of 
	    Nothing -> continue
	    Just p -> case getFragmentString p "message/sender" of
		Just s -> do
		    modifyFilter (FilterSender s:)
		    rcontinue
		Nothing -> continue
    (KeyChar 'c') -> do
	(ys,xs) <- scrSize
	v <- commandRead (getKey $ return ()) stdScr (ys - 1) "Filter: " ""
	case v of 
	    Just f -> maybe (return ()) (\x -> modifyFilter (x:)) $ parseFilter f 
	    Nothing -> setMessage ""
	rcontinue
    (KeyChar '~') -> do
	(ys,xs) <- scrSize
	v <- commandRead (getKey $ return ()) stdScr (ys - 1) "Filter: " "~"
	case v of 
	    Just f -> maybe (return ()) (\x -> modifyFilter (x:)) $ parseFilter f 
	    Nothing -> setMessage ""
	rcontinue
    (KeyChar 's') -> do
	(ys,xs) <- scrSize
	v <- commandRead (getKey $ return ()) stdScr (ys - 1) "Filename: " ""
	setMessage "feature coming soon!"
	rcontinue
    (KeyChar 'S') -> do
	(ys,xs) <- scrSize
	v <- commandRead (getKey $ return ()) stdScr (ys - 1) "Filename: " ""
	setMessage "feature coming soon!"
	rcontinue
    (KeyChar 'f') -> do
	p <- sel_puff
	case p of
	    Nothing -> continue
	    Just p -> composePuff (unwords $ map catShowNew $ cats p) >> rcontinue
    (KeyChar 'p') -> composePuff "" >> rcontinue
    (KeyChar 'r') -> do
	p <- sel_puff
	case p of
	    Nothing -> continue
	    Just p -> composePuff (getAuthor p) >> rcontinue
    (KeyF 1) -> (withLock screenLock helpScreen) >> hcontinue
    (KeyChar '\ESC') -> (withLock screenLock helpScreen) >> hcontinue
    (KeyChar n) | n `elem` "123456789" -> gfs (ord n - ord '1') >> rcontinue
    (KeyChar key) -> setMessage ("Unknown Key: " ++ show key  ++ " -- " ++ helpText) >> rcontinue
    key-> setMessage ("Unknown Key: " ++ show key  ++ " -- " ++ helpText) >> rcontinue
    
    where
	setMessage m = writeIORef message_r m
	getKey action = readChan ic >>= \v -> case v of
	    (Right p) -> addPuff p >> action >> getKey action
	    (Left k) -> return k
	continue = mainLoop gc ic yor psr selected_r filter_r next_r message_r cache_render_r cache_filter_r mark_set_r
	rcontinue = redraw >> continue
	hcontinue = (getKey $ return ()) >> rcontinue
	redraw = withLock screenLock $ (erase >> doRedraw >> refresh)
	mcontinue = readChan ic >>= \x -> case x of
	    (Right p) ->  addPuff p >> redraw >> mcontinue
	    (Left (KeyChar n)) | n `elem` "123456789" -> do
		fs <- readIORef filter_r
		modifyIORef mark_set_r (// ([(ord n - ord '1',fs)])) >> continue
	    (Left (KeyChar key)) -> setMessage ("invalid mark: " ++ show key) >> rcontinue
	    (Left key)-> setMessage ("invalid mark: " ++ show key ) >> rcontinue
	scrollPuffs x = modifyIORef yor (+ x) >> clip_yo 
	gfs n = do
	    mark_set <- readIORef mark_set_r
	    modifyFilter (const (mark_set ! n))
	filter_thread = do
	    p <- sel_puff
	    case p of 
		Just (Puff {cats = [c]}) -> modifyFilter (FilterCategory c:)
		Just (Puff {cats = cs}) -> modifyFilter (FilterOr (map FilterCategory cs):)
		_ -> return ()
	sel_puff = do
	    ps <- readIORef psr
	    selected <- readIORef selected_r
	    return $ case lookup selected ps of
		(Just p) | selectable p -> Just p
		_ -> Nothing
	addPuff p = do
	    n <- readIORef next_r
	    modifyIORef next_r (+1)
	    modifyIORef psr ((n,p):) 
	    writeIORef cache_render_r ([], -1) 
	    case p of 
		(Message m) -> setMessage m
		_ -> return ()
	    select_perhaps select_next
	getRendered = do
	    (ys,xs) <- scrSize
	    (cr,xs') <- readIORef cache_render_r
	    if xs == xs' then return cr else do
		ps <- readIORef psr 
		fs <- readIORef filter_r
		let f p = all (`filterAp` p) fs
		let nps =  [(x,rPuff xs p)| x@(_,p) <- ps] 
		writeIORef cache_render_r ((nps, xs)) 
		writeIORef cache_filter_r [x| x@((_,p),_) <- nps, f p]
		return nps
	composePuff cs = do
	    e <- getEditor
	    gs <- configLookupElse "GSEND" (return "gsend")
	    fn <- getTmpFile
	    nfn <- getTmpFile
	    let it = ["To: " ++ cs, "--------", ""] 
	    writeFile fn (unlines it)
	    def_prog_mode
	    endWin
	    mySystem (e ++ " +3 " ++ fn)
	    pn <- fmap lines $ readFile fn 
	    writeFile nfn (unlines $ (drop 2 pn))
	    if (length pn > 1 && pn /= it) then mySystem (gs ++ " -t 'id/class=" ++ fullName ++ "' " ++ drop 4 (head pn) ++ " < " ++ nfn) >> return () else 
		setMessage "Puff aborted."
	    reset_prog_mode
	modifyFilter f = do
	    modifyIORef filter_r f 
	    fs <- readIORef filter_r
	    setMessage (unwords (reverse $ map showFilter fs))
	    nps <- getRendered
	    let f p = all (`filterAp` p) fs
	    writeIORef cache_filter_r [x| x@((_,p),_) <- nps, f p]
	    select_perhaps select_next >> select_perhaps select_prev
	    center_puff
	    clip_yo
	getFiltered = do
	    getRendered
	    readIORef cache_filter_r
	buf_size = getFiltered >>= \ps -> return $ sum [x| (_,(x,_)) <- ps]
	center_puff = do
	    (ys,xs) <- scrSize
	    selected <- readIORef selected_r
	    ps <- liftM reverse getFiltered
	    let f t (((n,_),(h,_)):_) | n == selected = (t,t+h)
		f t ((_,(h,_)):rest) = f (t + h) rest
		f _ [] = (1000000,0)
	    let (mn,mx) = (f 0) ps
	    modifyIORef yor (min mn)
	    modifyIORef yor (max (mx - (ys - 1) ))
	scroll_end = do
	    (ys,xs) <- scrSize
	    bs <- buf_size
	    writeIORef yor (max (bs - (ys - 1)) 0)
	clip_yo = do
	    (ys,xs) <- scrSize
	    bs <- buf_size
	    modifyIORef yor (max 0)
	    modifyIORef yor (min (max 0 (bs - (ys - 1))))
	doRedraw = do 
	    (ys,xs) <- scrSize
	    selected <- readIORef selected_r
	    ps <- getFiltered
	    yo <- readIORef yor
	    m <- readIORef message_r
	    let rp y [] =  return () 
		rp y (((i,_),(n,f)):rest) = do
		    when ((0,ys) `ovl` (y - yo, y - yo + n)) $ f stdScr (y - yo) (i == selected)
		    rp (y + n) rest
	    attempt (rp 0 (reverse ps))
	    -- draw status bar
	    wmove stdScr (ys - 1) 0
	    wClrToEol stdScr
	    standout >> waddstr stdScr (' ':m) >> standend
	    let s = show selected ++ "/" ++ show (length ps)
	    wmove stdScr (ys - 1) (xs - length s - 2)
	    wAttrOn stdScr attrBold >> waddstr stdScr s >> wAttrOff stdScr attrBold
	select_perhaps a = do
	    ps <- getFiltered
	    n <- readIORef selected_r
	    unless (maybe False selectable $ lookup n (fsts ps)) a
	select_prev = do
	    ps <- liftM fsts getFiltered
	    let f sel ((n,p):ps) | n < sel && selectable p = n 
		f sel (_:ps) = f sel ps
	        f sel [] = sel
	    modifyIORef selected_r (\s -> f s ps)
	select_next = do
	    ps <- liftM fsts getFiltered
	    let f sel bg ((n,p):ps) | n > sel && selectable p = f sel n ps
		f sel bg ((n,_):ps) | n > sel = f sel bg ps
	        f sel bg _ = bg
	    modifyIORef selected_r (\s -> f s s ps)



getSenderText :: Puff -> String
getSenderText p@(Puff {signature = sig, fragments = frags}) = case getFragmentString p "message/sender" of
	Just s -> s
	Nothing -> getAuthor p
	
getAuthor p = ss (signature p) where
	    ss (Signed (Key n _):_) = n
	    ss (Unverifyable (Key n _):_) = n
	    ss (_:sig) = ss sig
	    ss [] = "<unknown>"


rPuff :: Int -> Puff -> (Int, Window -> Int -> Bool -> IO ())
rPuff mw p@(Puff {cats =cats, fragments = frags}) = (n + 2,f) where
    reformat s | length s < mw = [s]
    reformat s = let (a,b) = splitAt mw s in a : reformat b
    sender = getSenderText p
    body = case lookup "message/body" frags of
	Just (FragmentText s) -> lines s
	_ -> []
    time = case lookup "id/time" frags of 
	Just (FragmentTime t) -> Just t
	_ -> Nothing
    kwds = getFragmentStrings p "message.keyword"
    body' = concatMap reformat body
    n = length body'
    addCat w (x,y) = wAttrOn w attrBold >> waddstr w x >> wAttrOff w attrBold >> waddstr w ('@':y) 
    f w y selected = do
	when (y >= 0) $ do
	    wmove w y 0
	    sequence (intersperse (space w) $ map (addCat w) cats)
	    space w >> space w
	    waddstr w (unwords (map ('/':) kwds))
	    st <- case time of 
		Just t -> liftM ft $ toCalendarTime t
		Nothing -> return ""
	    let lst = length st
	    wAttrOn w attrBold
	    mvwaddstr w y (mw - length sender - lst - 1) sender
	    wAttrOff w attrBold
	    when selected $ void standout
	    mvwaddstr w y (mw - lst) st
	    when selected $ void standend
	mapM_ (\(p,s) ->  mvwaddstr stdScr (y + p) 0 s) (zip [1..] $ body')
	wAttrOn w attrBold
	mvwaddstr w (y + 1 + n) 0 (replicate mw $ if selected then '=' else '-')
	wAttrOff w attrBold
	return ()
    ft (CalendarTime {ctHour = h, ctMin = m, ctSec = s}) = fn h ++ ":" ++ fn m ++ ":" ++ fn s
    fn n = let x = show n ; lx = length x in if lx < 2 then replicate (2 - lx) '0' ++ x else x
rPuff mw p = tup 1 $ \w y _ -> do
    wAttrOn w attrDim
    mvwaddstr w y 0 (showPuff p)
    wAttrOff w attrDim

tup x y = (x,y)
	
(a,b) `ovl` (x,y) | y < a = False
(a,b) `ovl` (x,y) | b < x = False
_ `ovl` _ = True


------------------
-- Curses routines
------------------

addstr s = Control.Exception.try (addStr s) >> return ()
waddstr w s = Control.Exception.try (wAddStr w s) >> return ()
mvwaddstr w y x s = Control.Exception.try (mvWAddStr w y x s) >> return ()
wmove w y x = Control.Exception.try (wMove w y x) >> return ()
cursset x = Control.Exception.try (cursSet x) >> return ()
addln = Control.Exception.try addLn >> return ()

space w = waddstr w " "

--------------
-- help screen
--------------

buildTable :: [(String,String)] -> [String]
buildTable ps = map f ps where
    f (x,y) = replicate (bs - length x) ' ' ++ x ++ replicate 4 ' ' ++ y
    bs = 2 + maximum (map (length . fst) ps)



helpTable = [
    ("Keys:", ""),
    ("j","next puff"),
    ("k","previous puff"),
    ("G","last puff"),
    ("g","first puff"),
    ("<enter> ^E","forward one line"),
    ("<backspace> ^Y","back one line"),
    ("^F <PgDown>","forward one page"),
    ("^B <PgUp>","back one page"),
    ("^D","forward one half-page"),
    ("^U","back one half-page"),
    ("",""),
    ("c ~","add new filter"),
    ("u","pop one filter"),
    ("U","pop all filters"),
    ("!","invert filter at top of stack"),
    ("x","swap top two filters on stack"),
    ("a","filter to current author"),
    ("t","filter to current thread"),
    ("T","filter to strict thread"),
    ("m[1-9]","set mark - save current filter stack at given key"),
    ("[1-9]","recall filter stack saved here with 'm'"),
    ("",""),
    ("f","follow-up to category"),
    ("r","reply to author"),
    ("p","compose new puff"),
    ("",""),
    ("s","save selection to file"),
    ("S","save current puffs to file"),
    ("^R","reconnect to all servers"),
    ("E","edit and reload configuration file"),
    ("q","quit program"),
    ("^L","redraw screen"),
    ("",""),
    ("Filters:",""),
    ("~a","author"),
    ("~s","sender"),
    ("~c","category"),
    ("(default)","substring"),
    ("~k","keyword"),
    ("~t","thread - transitive closure of categories")
    ]

helpScreen = do
    (ys,xs) <- scrSize
    let s = "Help Screen - press any key to exit."
    erase
    mvwaddstr stdScr 0 ((xs - length s) `div` 2 ) s
    move 2 0
    mapM_ (\s -> addstr s >> addln) (buildTable helpTable)

    refresh
    
helpText = "Need Help? Press F1"

-------------------------
-- options/initialization
-------------------------

env = Env { 
    envAction = return (),
    envVerbose = False, 
    envConfig = Nothing,
    envJustArgs = False
    } 


options :: [OptDescr (Env -> Env)]
options = [
	Option ['v']  ["verbose"] (NoArg (\e->e{envVerbose = True})) "verbose output on stderr",
	Option ['V']  ["version"] (NoArg (doAction doVersion)) "print version information",
--	Option ['c']  ["config"] (ReqArg (\x e->e{envConfig = (Just x)}) "FILE") "use this configuration file",
	Option ['s']  ["sample"] (NoArg (doAction doSample)) "print sample configuration file to stdout",
	Option ['m']  ["man"] (NoArg (doAction doMan)) "print all internal help screens to stdout",
	Option ['e']  [] (NoArg (\e->e{envJustArgs = True})) "only subscribe to command line arguments"
    ]

void a = a >> return ()

ginsuOpts :: IO (Env,[String])
ginsuOpts = do
    args <- getArgs
    r@(env,_) <- case (getOpt Permute options args) of
	   (as,n,[]) -> return (foldr ($) env as ,n)
    	   (_,_,errs) -> putErrDie (concat errs ++ usage)
    envAction env
    handle (\_ -> putErrDie ("Could not find configuration file: ~/.gale/ginsu.config\n" ++ trailer)) (loadStdConfiguration ".gale/ginsu.config")
    return r

doVersion = putStrLn fullName >> exitSuccess
doSample = putStr exampleConf >> exitSuccess
doMan = putStrLn usage >> putStrLn (unlines (buildTable helpTable)) >> exitSuccess

usage = usageInfo header options ++ trailer where
    header = "Usage: ginsu [OPTION...] categories..."
trailer = "\nTo get started, generate a configuration file using the following command:\nmkdir ~/.gale ; ginsu -s > ~/.gale/ginsu.config\nEdit the configuration file and run ginsu!\n\n"
 ++ "The homepage is at <http://repetae.net/john/computer/ginsu/>\nTo report a bug go to <http://bugs.ofb.net/cgi-bin/bugzilla/enter_bug.cgi?product=Ginsu>\nWhen reporting a bug, include the last few lines of ~/.gale/ginsu.errorlog if it seems appropriate."

doAction :: IO a -> (Env -> Env)
doAction a = \e -> e {envAction = a >> envAction e}



commandRead :: IO Curses.Key -> Window -> Int -> String -> String -> IO (Maybe String)
commandRead ic win yloc prompt init = do
    cursset Visible
    v <- cr (reverse init) 
    cursset Invisible
    return v
    where
    cr v = pc v >> ic >>= \x -> case x of
	KeyEnter -> return $ Just (reverse v)
	(KeyChar '\n') -> return $ Just (reverse v)
	(KeyChar '\r') ->  return $ Just (reverse v)
	KeyBackspace -> cr (drop 1 v)
	(KeyChar c) -> cr (c:v)
	_ -> return Nothing
    pc v = do
	wmove win yloc 0
	wClrToEol win
	wAttrOn win attrBold
	waddstr win prompt 
	wAttrOff win attrBold
	waddstr win (reverse v)
	wmove win yloc (length v + length prompt)
	refresh
